/*
 *
 * Copyright (C) 2003 Alexander Yip (yipal@lcs.mit.edu)
 *
 *  Permission is hereby granted, free of charge, to any person obtaining
 *  a copy of this software and associated documentation files (the
 *  "Software"), to deal in the Software without restriction, including
 *  without limitation the rights to use, copy, modify, merge, publish,
 *  distribute, sublicense, and/or sell copies of the Software, and to
 *  permit persons to whom the Software is furnished to do so, subject to
 *  the following conditions:
 *
 *  The above copyright notice and this permission notice shall be
 *  included in all copies or substantial portions of the Software.
 *
 *  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
 *  EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
 *  MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
 *  NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE
 *  LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION
 *  OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION
 *  WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
 *
 */
#include <misc_utils.h>
#include <configurator.h>

#include "chord.h"
#include "dhash.h"
#include "dhashgateway.h"
#include "dhashclient.h"
#include "dhash_common.h"
#include "sfscrypt.h"
#include "parseopt.h"
#include <sys/types.h>
#include "route.h"
#include "crypt.h"
#include "verify.h"
#include <dhash_prot.h>
#include <control_prot.h>
#include <admcontrol.h>

#include <debruijn.h>
#include <fingerroute.h>
#include <fingerroutepns.h>
#include <proxroute.h>
#include <recroute.h>

#include <modlogger.h>
#define info modlogger ("lsd")

dhashclient *dhc;
ptr<sfspriv> skwriter;
ptr<sfspriv> skproject;
ptr<sfspriv> skaqua;
ptr<sfs_pubkey2> pkw; 
ptr<sfs_pubkey2> pkp; 

int rpcs_out;

str
sig2str (sfs_sig2 *sig) {
  xdrsuio x;
  if (!xdr_sfs_sig2 (&x, sig))
    return NULL;
    
  char *m_sig = suio_flatten(x.uio());
  return armor64(m_sig, x.uio()->resid());
}

str
pubkey2str (sfs_pubkey2 key)
{
  strbuf pkstrbuf;
  ptr<sfspub> pubkey = sfscrypt.alloc(key, SFS_VERIFY);
  pubkey->export_pubkey (pkstrbuf, false);
  return str(pkstrbuf);
}

void
generate_keys(str label) {
  str s;
  strbuf sb;
  ptr<sfspriv> sk_aqua;

  sk_aqua = sfscrypt.gen(SFS_RABIN, 0, SFS_SIGN);
  sk_aqua->export_privkey (&s, label, NULL, 0);
  str2file (str ( strbuf() << label << ".sk"), s);

  sk_aqua->export_pubkey(sb);
  str2file (str( strbuf() << label << ".pk"), str(sb));
}

bool
sign_hostlist(str filename, str keyfile)
{
  str contents, rawkey;
  if ((contents = file2str(filename)) &&
      (rawkey = file2str (keyfile))) {

    sfs_sig2 sig;
    ptr<sfspriv> aqua_sk = sfscrypt.alloc_priv (rawkey, SFS_SIGN);
    aqua_sk->sign(&sig, contents);

    strbuf signedlist;
    signedlist << sig2str(&sig) << "\n" << contents;
    //bool r = str2file("signedhostlist", signedlist);
    //warn << "Signed hostlist, created signedhostlist: " << r << "\n";
    printf("%s", str(signedlist).cstr());
    return true;

  } else {
    warn << "could not sign " << filename<< "\n";
  }
  return false;
}

void
dist_send_cb (ptr<control_dist_arg> arg, control_dist_res *res, clnt_stat err) 
{
  warn << "sent contents: " << err << "\n";
  if (res) {
    warn << "status: " << res->status << "\n";
    delete res;
  }

  if (--rpcs_out == 0) 
    exit(0);
}

ptr<aclnt> clnt;

void
dist_connect_cb (str contents, int fd) 
{
  if (fd < 0)
    return;

  ptr<control_dist_arg> arg = New refcounted<control_dist_arg>();
  arg->hostlist = true;
  arg->contents.setsize(contents.len());
  memcpy (arg->contents.base(), contents.cstr(), contents.len());

  control_dist_res *res = New control_dist_res;

  clnt = aclnt::alloc (axprt_stream::alloc(fd), control_program_1);
  clnt->call (CONTROLPROC_DIST, arg, res, wrap(&dist_send_cb, arg, res));
  rpcs_out++;
}

void
dist_hostlist(str filename) 
{
  rpcs_out = 0;
  str signedhostlist = file2str(filename);
  if (!signedhostlist) {
    warn << "could not read hostlist\n";
    exit(1);
  }

  int version;
  vec<chord_hostname> names;
  vec<int32_t> ports;
  if (1 == hostlist::parse_hostlist(signedhostlist, &version, &names, &ports, NULL, NULL)) {
    for(unsigned i=0; i<names.size(); i++){
      // XXX dont do all of them at the same time. 
      //     (wellknown might be down during restart)

      warn << "sending hostlist to " << names[i] << " " << ports[i] << "\n";
      tcpconnect (names[i], ports[i]+3, wrap(&dist_connect_cb, signedhostlist));
    }
  } else {
    warn << "could not parse hostlist\n";
    exit(1);
  }

  amain();
}

static void
usage ()
{
  warnx << "Usage: " << "pastwatchtest [-g keylabel] [-k skey -s file] [-h hostlist]\n"
	<< " -g generate keys\n"
	<< " -k the secret key file\n"
	<< " -s sign file and output to stdout\n"
	<< " -h distribute hostlist to nodes\n"
	<< "\n";
  exit (1);
}

int
main (int argc, char **argv)
{
  int ch;
  str keyfile, signfile;
  if (argc < 2) 
    usage();

  while((ch = getopt(argc, argv, "g:k:s:h:")) != -1) {
    switch(ch) {
    case 'g':
      generate_keys(str(optarg));
      break;
    case 'k':
      keyfile = str(optarg);
      break;
    case 's':
      signfile = str(optarg);
      break;
    case 'h':
      dist_hostlist(str(optarg));
      break;
    default:
      usage();
      break;
    }
  }

  if (signfile && keyfile) {
    sign_hostlist(signfile, keyfile);
  }

  return 0;
}
