/*
 *
 * Copyright (C) 2004 Alexander Yip (yipal@lcs.mit.edu)
 *
 *  Permission is hereby granted, free of charge, to any person obtaining
 *  a copy of this software and associated documentation files (the
 *  "Software"), to deal in the Software without restriction, including
 *  without limitation the rights to use, copy, modify, merge, publish,
 *  distribute, sublicense, and/or sell copies of the Software, and to
 *  permit persons to whom the Software is furnished to do so, subject to
 *  the following conditions:
 *
 *  The above copyright notice and this permission notice shall be
 *  included in all copies or substantial portions of the Software.
 *
 *  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
 *  EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
 *  MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
 *  NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE
 *  LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION
 *  OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION
 *  WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
 *
 */

#include "gcc_compat.h"
#include <unistd.h>
#include <sys/types.h>
#include <sys/socket.h>
#include "sfsmisc.h"
#include "arpc.h"
#include "async.h"
#include "str.h"
#include "controlserver.h"
#include "control_prot.h"
#include <blacklist.h>

controlserver::controlserver(str name, int port, ptr<axprt_stream> x, ptr<restarter> r):
  myname(name), myport(port), _restarter(r)
{
  clntsrv = asrv::alloc(x, control_program_1, wrap( mkref(this), &controlserver::dispatch));
}
controlserver::~controlserver() {
}

void
controlserver::dispatch(svccb *sbp) {
  if (!sbp) {
    clntsrv = 0;
    return;
  } 


  switch(sbp->proc()) {
  case CONTROLPROC_DIST:
    {
      control_dist_arg *arg = sbp->TEMPLATE getarg<control_dist_arg> ();
      str contents = str(arg->contents.base(), arg->contents.size());
      
      int newversion = hostlist::only().verify_hostlist(contents);
      control_dist_res *res;
      
      if (newversion < 0) {
	res = New control_dist_res(CONTROL_DIST_NOVERIFY);
	sbp->reply(&res);
	delete res;
	return;
      } 

      if (newversion > hostlist::only().get_version()) {
	warn << "Updating hostlist:  " << hostlist::only().get_version()
	     << " -> " << hostlist::only().verify_hostlist(contents) << "\n";

	res = New control_dist_res(CONTROL_OK);
	sbp->reply(&res);
	delete res;
      
	hostlist::only().write_hostlist(contents);
	delaycb(10, wrap(mkref(this), &controlserver::restart));
	return;
      }
    }
  case CONTROLPROC_SYNC:
    {
      control_sync_arg *arg = sbp->TEMPLATE getarg<control_sync_arg>();

      if (hostlist::only().get_version() < arg->hostlistver) {
	ptr<fetchwrap> fw = New refcounted<fetchwrap>();
	fw->arg = New refcounted<control_fetch_arg>();
	fw->res = New refcounted<control_fetch_res>();
	fw->arg->hostlist = true;
	fw->arg->ip = 0;
	fw->arg->port = 0;

	// All fetches keep a reference to this object, so that 
	// this does not get freed before the responses come back
	tcpconnect(arg->caller_name, arg->caller_port+3, 
		   wrap(mkref(this), &controlserver::send_fetch, fw));
      }


      // compare my blacklist versions to remote versions
      // if a remote blacklist is higher than local version, send RPC asking for it.
      
      control_sync_res res(CONTROL_OK);
      res.resok->resp_name = myname;
      res.resok->resp_port = myport;
      res.resok->hostlistver = hostlist::only().get_version();

      hostentry *p = hostlist::only().first();
      res.resok->node_ips.setsize(hostlist::only().size());
      res.resok->node_ports.setsize(hostlist::only().size());
      res.resok->versions.setsize(hostlist::only().size());
      
      int i = 0;
      while(p) {
	res.resok->node_ips[i] = p->name;
	res.resok->node_ports[i] = p->port;
	res.resok->versions[i] = blacklist::only().getversion(p->name, p->port);
	i++;
	p = hostlist::only().next(p);
      }

      sbp->reply(&res);
      break;
    }
  case CONTROLPROC_FETCH:
    {
      control_fetch_arg *arg = sbp->TEMPLATE getarg<control_fetch_arg>();
      control_fetch_res res(CONTROL_OK);

      if(arg->hostlist) {
	str hl = hostlist::only().get_raw_hostlist();
	res.resok->contents.setsize(hl.len());
	memcpy(res.resok->contents.base(), hl.cstr(), hl.len());
      } else {
	// XXX place the blacklist contents in reply
      }

      sbp->reply(&res);
      break;
    }
  default:
    warn << "Unknown RPC\n";
    control_stat res = CONTROL_ERR;
    sbp->reply(&res);
  }
}

void
controlserver::start_listen(str name, int port, int listenport, ptr<restarter> r) {
  int fd = inetsocket (SOCK_STREAM, listenport);
  if (fd < 0) 
    fatal << "Error creating control socket " << listenport << " (TCP) " << strerror(errno) << "\n";

  close_on_exec(fd);

  if (listen(fd, 5) < 0) {
    fatal ("Error from listen: %m\n");
    close(fd);
  } else {
    r->add_port(fd);

    ptr<argwrap> a = New refcounted<argwrap>();
    a->myname = name;
    a->myport = port;
    a->fd = fd;
    a->restart = r;

    fdcb (fd, selread, wrap(&controlserver::accept_control_client, a, fd));
  }
}

void
controlserver::accept_control_client(ptr<argwrap> a, int lfd) {
  sockaddr_un sun;
  bzero (&sun, sizeof(sun));
  socklen_t sunlen = sizeof(sun);
  int fd = accept (lfd, reinterpret_cast<sockaddr *> (&sun), &sunlen);

  if (fd < 0) {
    warn << "Could not accpet controlserver client.\n";
    return;
  }

  ref<axprt_stream> x = axprt_stream::alloc(fd, 1024*1025);
  vNew refcounted<controlserver> (a->myname, a->myport, x, a->restart);
}

void
controlserver::send_fetch(ptr<fetchwrap> fw, int fd) {
  if (fd < 0) {
    warn << "could not connect to fetch new lists\n";
    return;
  }
  
  fw->clnt = aclnt::alloc (axprt_stream::alloc(fd), control_program_1);
  fw->clnt->call (CONTROLPROC_FETCH, fw->arg, fw->res, 
		  wrap(mkref(this), &controlserver::send_fetch_cb, fw ));

}

void
controlserver::send_fetch_cb(ptr<fetchwrap> fw, clnt_stat err) {
  str newhostlist(fw->res->resok->contents.base(), fw->res->resok->contents.size());

  warn << "send_fetch_cb: " << err << "\n";
  warn << "  contents: " << newhostlist << "\n";

  int newversion = hostlist::only().verify_hostlist(newhostlist);
  if (newversion > hostlist::only().get_version()) {
    hostlist::only().write_hostlist(newhostlist);
    delaycb(10, wrap(mkref(this), &controlserver::restart));
  }

  return;
}

void
controlserver::restart() {
#if 0
  warn << "closing listen sockets\n";
  //lp->close_all();
  sleep(3);
  warn << "restarting\n";
  execv(argv[0], argv);
#endif

  
}



