/*
 *  Copyright (C) 2002-2004  Massachusetts Institute of Technology
 *
 *  This program is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU General Public License as
 *  published by the Free Software Foundation; either version 2, or (at
 *  your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful, but
 *  WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *  General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307
 *  USA
 */

#ifndef REPLICA_IMPL_H
#define REPLICA_IMPL_H

#include "stdinc.h"
#include "logw.h"
#include "cfsr.h"
#include "replica.h"
#include "snapshot.h"
#include "btree.h"
#include "pastdbfe.h"

typedef callback<void, bigint, cfsrw_status>::ptr cbaddr;
typedef vec<sptr> spv;

struct replica_meta
{
  vervec pkv;         // Maintains public key block versions (from ns.C)
  comvec commits;     // Current list of log pointers, and their corresponding log heads
  vec<bigint> leaves; // List of leaves (the addresses of their snapshots)

  void serialize (strbuf &sb) const;
  bool parse (xin &x);
};

class replica_impl : public replica
{
private:
  struct replica_meta _m;
  bigint _projid;
  bigint _replicaaddr;
  bool _changed;
  ptr<pastdbfe> _rdb;
  ptr<btree_srv> _btree;

  spv _branches; // all the current branch heads for the loaded portion of the branch tree
  spv _slist;    // all the snapshots we ever saw
  str _mybptag;
  str _mybmagic;
  int _mybranch;

  void set_addrs (spv);
  void load_cb (cbloadsn_t cb, sptr s, cfsrw_status status);
  void load_next (vec<bigint> a, cbcfsrw cb, sptr s, cfsrw_status status);
  void build_cb (comvec commits, cbcfsrw cb, spv v, cfsrw_status status);
  void find (bigint saddr, cbloadsn_t cb);
  void dispatch_with_snapshot (cfsrcallp c, sptr s, cfsrw_status status);

  ptr<pastdbfe> open_rdb ();

public:
  replica_impl (bigint v, bigint r);
  virtual ~replica_impl () {
    exit_cleanup ();
  }
  void exit_cleanup () {
    if (_rdb)
      _rdb->closedb ();
    _rdb = 0;
  }
  const vervec& pkv () const { return _m.pkv; }
  const comvec& commits () const { return _m.commits; }

  void load (cbcfsrw cb);
  void update (comvec head, cbcfsrw cb);
  void sync (cbaddr cb);
  bool changed () const { return _changed; }
  void setpkv (const vervec& v) {
    if (_m.pkv.equals (v))
      return;
    _changed = true;
    _m.pkv = v;
  }

  void dispatch (cfsrcallp c);
  bool lookup_snapshot (bigint, bigint&);
 
  // public interface exposed by parent class replica

  bigint root () const;
  unsigned nbranches () const { return _branches.size (); }

  str branch_tag (unsigned) const;
  bigint branch_head_snapshot (unsigned) const;
  commit_info branch_head_info (unsigned) const;
  int my_branch () const;

  bool set_branch_head (bigint);
  int get_branch_index (bigint n) const;
  ptr<btree_srv> btree () const { return _btree; }
  bigint determine_current_branch(str arg_branch, bool ignore_implicit_branches, 
				  str crec_from_info) const;

  int map_crec_to_snapaddr(bigint crecord, bigint snapshot);
  bigint crec_to_snapaddr (bigint crecord);
};

#endif

