/*
 *  Copyright (C) 2002-2004  Massachusetts Institute of Technology
 *
 *  This program is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU General Public License as
 *  published by the Free Software Foundation; either version 2, or (at
 *  your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful, but
 *  WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *  General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307
 *  USA
 */

#include "parseopt.h"
#include "stdinc.h"
#include "projinterface.h"
#include "cfsrw.h"
#include "replica.h"
#include "btree.h"
#include "cfsr.h"
#include "cfsw_record.h" // cfsw_deltap_parse

#include "pastwatch.h"
#include "opts.h"
#include "pdb.h"
#include "path.h"
#include "io.h"
#include "misc_utils.h"

#include "sfskeymgr.h"

u_int64_t t0;

ptr<ns> nsobj;
str _myname;
ptr<cfsrw> _cfsrw;
bigint _root = 0;
bool command_done_called = false;
bool anonymous = false;
static ptr<sfspriv> sk;

pdb_mgr* pmgr = 0;
static ptr<vec<str> > files = 0;

extern vec<str> notify_cmds;
extern vec<str> notify_files;

extern str get_btag_from_infofile (str path);
extern str get_branch_from_infofile (str path);
extern str get_commit_record_from_infofile (str path);
extern str translate_key_to_name (str t);
extern str translate_branch_tag (str t);
extern void list_branches (int err = 0);
extern void
find_commit_by_time (unsigned t, bigint c, callback<void,bigint,int>::ref cb);
extern void create_new_log (callback<void, log_status>::ptr cb);


extern "C" {
time_t get_date (char *raw);
};

static void translate_time (bigint c);

static inline int
reset_global_states ()
{
  notify_cmds.clear ();
  notify_files.clear ();
  command_done_called = false;
  pmgr = New pdb_mgr ();
  return pmgr->init ();
}

static inline bool
op_need_repository (int command)
{
  if (command == cmd_clear)
    return false;
  return true;
}

//
// driver procedures in reverse order; main is at the end of the file,
// and the last procedure to run is below.
//

static void nextproj1 ();
static void nextproj2 ();
static void nextproj3 (log_status status);

static void
pastwatch_exit (int n)
{
  if (n) {
    if (_cfsrw) {
      if (_cfsrw->getreplica ())
	_cfsrw->getreplica ()->exit_cleanup ();
      _cfsrw = 0;
    }
    if (nsobj) {
      nsobj->exit_cleanup ();
      nsobj = 0;
    }
    exit (n);
  }
  else
    nextproj1 ();
}

static void
notify_done (int error)
{
  if (error)
    warn << "unable to notify peers, continuing.\n";
  pastwatch_exit (0);
}

extern void notify (cbi cb);

static void
update_local (int error)
{
  if (!error) {
    if (!arg_checkonly) {
      assert (arg_branch != "");
      bigint cr_addr = my_branch_head (_cfsrw->getreplica ());
      str btag =
	_cfsrw->getreplica ()->branch_tag
	(_cfsrw->getreplica ()->my_branch ());
      pmgr->process (cr_addr, btag);
  
      // u_int64_t d = getusec () - t0;
      // warn << "process: +" << (long)d << "\n";
    }
  }
  if (!error)
    notify (wrap (notify_done));
  else
    pastwatch_exit (error);
}

static void
commit_reply (int error, cfswcallp c, cfsrw_status status)
{
  // u_int64_t d = getusec () - t0;
  // warn << "commit: +" << (long)d << "\n";
  
  sigcb (SIGHUP, wrap (pastwatch_exit, -1));
  sigcb (SIGUSR1, wrap (pastwatch_exit, -1));
  sigcb (SIGUSR2, wrap (pastwatch_exit, -1));
  sigcb (SIGALRM, wrap (pastwatch_exit, -1));
  sigcb (SIGINT, wrap (pastwatch_exit, -1));
  sigcb (SIGTERM, wrap (pastwatch_exit, -1));

  if (status) {
    if (!error)
      warn << "warning: cannot commit (" << (int) status << "), continuing.\n";
    error = -1;
  } else if (_cfsrw->getreplica ()->my_branch () >= 0){
    arg_branch = armor64A (my_branch_head (_cfsrw->getreplica ()));
  } else{
    arg_branch = "";
  }
  update_local (error);
}

static void
abort (int error)
{
  if (!arg_readonly) {
    cfswcallp c = New refcounted<cfswcall>
      (CFSWPROC_ABORT, wrap (commit_reply, error));
    cfsw_call (c);
  }
  else
    commit_reply (error, 0L, CFSRW_OK);
}

static void
commit (int error)
{
  if (!arg_readonly && !anonymous) {
    warn << "committing\n";

    cfswcallp c = New refcounted<cfswcall>
      (CFSWPROC_COMMIT, wrap (commit_reply, error));
    cfsw_commit_arg *arg = static_cast<cfsw_commit_arg *> (c->arg ());
    arg->message = arg_message;
    if (arg_branchtag && arg_branchtag != "")
      arg_branchtag = _myname << ":" << arg_branchtag;
    arg->btag = arg_branchtag;
    arg->flags = 0;
    if (arg_timestamp && arg_timestamp != "") {
      time_t t = get_date ((char*) arg_timestamp.cstr());
      arg->commit_time = t;
    }

    if (arg_newbranch) {
      arg->flags |= commit_record::BRANCH_OFF;
      bigint f;
      if (arg_branchfrom && arg_branchfrom != "")
        dearmor64A_bigint (arg_branchfrom, f);
      else
        f = my_branch_head (_cfsrw->getreplica ());
      arg->fork = f;
    }

    else if (arg_killbranch)
      arg->flags |= commit_record::IGNORE_BRANCH;

    else if (arg_jointag && arg_jointag != "") {
      bigint j;
      dearmor64A_bigint (arg_jointag, j);
      arg->flags |= commit_record::MERGE_WITH;
      arg->supersede = j;
    }

    cfsw_call (c);
  }
  else
    commit_reply (error, 0L, CFSRW_OK);
}

static void dispatch ();

static void
command_done (int error)
{
  // u_int64_t d = getusec () - t0;
  // warn << "dispatch: +" << (long)d << "\n";

  // if there is an error, we abort all the changes we made propagate
  // the error, so we don't change the local version db.  
  //
  // NOTE: can't call abort unless already in async mode; otherwise we
  // may corrupt the log.

  if (command_done_called)
    return;
  command_done_called = true;

  if (error)
    warn << "error/interrupted while running "
         << cmd2str (arg_command) << "\n";

  if (error && !arg_readonly)
    abort (error);
  else if (error || files->empty ())
    commit (error);
  else
    dispatch ();
}

extern void update (bigint root, str path, cbi cb);
extern void diff (bigint root, str path, cbi cb);
extern void commit (bigint root, str path, cbi cb);
extern void import (bigint root, str path, cbi cb);
extern void checkout (bigint root, str path, str cfspath, cbi cb);
extern void rescue (bigint root, str path, bigint handle, cbi cb);
extern void log (bigint root, str, cbi);
extern void rename (bigint root, str from, str to, cbi cb);
extern void status (str, cbi);
extern void remove (bigint root, str path, cbi);
extern void add (bigint root, str path, cbi);
extern void rtag (str, cbi);
extern void admin (cbi);

static void
noop_cb (cbi cb, int n)
{
  cb (n);
}

static void
dispatch ()
{
  str p = files->pop_front ();
  command_done_called = false;

  if (arg_command == cmd_import)
    import (_root, p, wrap (command_done));
  
  else if (arg_command == cmd_commit)
    commit (_root, p, wrap (command_done));

  else if (arg_command == cmd_rename)
    rename (_root, p, files->pop_front (), wrap (command_done));

  else if (arg_command == cmd_add)
    add (_root, p, wrap (command_done));

  else if (arg_command == cmd_remove)
    remove (_root, p, wrap (command_done));

  else if (arg_command == cmd_rescue)
    rescue (_root, p, arg_handle, wrap (command_done));

  else if (arg_command == cmd_checkout)
    checkout (_root, arg_localpath, p, wrap (command_done));

  else if (arg_command == cmd_merge)
    update (_root, p, wrap (command_done));

  else if (arg_command == cmd_update)
    update (_root, p, wrap (command_done));

  else if (arg_command == cmd_diff)
    diff (_root, p, wrap (command_done));

  else if (arg_command == cmd_status)
    status (p, wrap (command_done));

  else if (arg_command == cmd_log)
    log (_root, p, wrap (command_done));
  
  else if (arg_command == cmd_rtag)
    rtag (p, wrap (command_done));

  else if (arg_command == cmd_latest ||
           arg_command == cmd_tags ||
	   arg_command == cmd_history)
    admin (wrap (command_done));
  
  else if (arg_command == cmd_branch ||
           arg_command == cmd_fetch)
    command_done (0);
  
  else
    assert (0);
}

static void lock_repository (bigint root);

static void
lock_reply (cfswcallp c, cfsrw_status status)
{
  // u_int64_t d = getusec () - t0;
  // warn << "lock: +" << (long)d << "\n";
  const ptr<replica> r = _cfsrw->getreplica ();

  if (status == CFSRWERR_EXCLCONT) {
    warn << "cannot acquire lock on repository\n";
    // don't need to cancel because new lock is never set
    pastwatch_exit (1);
    return;
  }
  else if (status == CFSRWERR_NOBRANCH) {
    bigint c = 0;
    if (r->nbranches () == 1) {
      c = r->branch_head_info (0).rptr;
      arg_branch = armor64A (c);
    }
    if (!r->set_branch_head (c)) {
      warn << "cannot find a branch to use. current branches are (A):\n";
      list_branches (1);
      pastwatch_exit (1);
      return;
    }
    // need to restart
    lock_repository (_root);
    return;
  }
  else if (status) {
    warn << "cannot get lock on repository (" << status << ")\n";
    pastwatch_exit (1);
    return;
  }

  str info_crec =  (files->size() > 0) ? get_commit_record_from_infofile((*files)[0]) : str("");
  bigint c2 = r->determine_current_branch(arg_branch, arg_ignore_implicit_branches, info_crec);
  r->set_branch_head (c2);

  if (_cfsrw->getreplica ()->my_branch () >= 0) {
    arg_branch = armor64A (my_branch_head (_cfsrw->getreplica ()));
    warn << "Using branch head: " << armor64A(c2) << "\n";
  } else {
    if (arg_command != cmd_latest && arg_command != cmd_fetch) {
      warn << "cannot find a branch to use. current branches are (B):\n";
      list_branches (1);
      pastwatch_exit (1);
      return;
    }
    arg_branch = "";
  }

  translate_time(c2);
}

static void
dispatch_local_one (str path, cbi cb)
{
  if (arg_command == cmd_clear) {
    // for "clear", just remove the entry from db
    pmgr->remove (path);
    str btag;
    if (_cfsrw->getreplica ()->my_branch () >= 0)
      btag =
	_cfsrw->getreplica ()->branch_tag
	(_cfsrw->getreplica ()->my_branch ());
    pmgr->process (0, btag);
    delaycb (0, wrap (noop_cb, cb, 0));
  }
  else
    assert (0);
}

static void
dispatch_local (int err)
{
  if (err) {
    pastwatch_exit (err);
    return;
  }
  else {
    if (files->empty ()) {
      pastwatch_exit (0);
      return;
    }
    else {
      str p = files->pop_front ();
      dispatch_local_one (p, wrap (dispatch_local));
    }
  }
}

static void
lock_repository (bigint root)
{
  // lock_repository retrieves new commtis from the central replica
  // and builds snapshots for the new commit records.

  //
  // run commands that do not require the repository
  //
  if (!op_need_repository (arg_command)) {
    delaycb (0, wrap (dispatch_local, 0));
    return;
  }

  sigcb (SIGHUP, wrap (command_done, -1));
  sigcb (SIGUSR1, wrap (command_done, -1));
  sigcb (SIGUSR2, wrap (command_done, -1));
  sigcb (SIGALRM, wrap (command_done, -1));
  sigcb (SIGINT, wrap (command_done, -1));
  sigcb (SIGTERM, wrap (command_done, -1));

  if (arg_readonly || anonymous) {
    cfswcallp c = New refcounted<cfswcall> (CFSWPROC_RLOCK, wrap (lock_reply));
    cfsw_rlock_arg *arg = static_cast<cfsw_rlock_arg *> (c->arg ());
    arg->handle = root;
    cfsw_call (c);
  }
  else {
    cfswcallp c = New refcounted<cfswcall> (CFSWPROC_WLOCK, wrap (lock_reply));
    cfsw_wlock_arg *arg = static_cast<cfsw_wlock_arg *> (c->arg ());
    arg->handle = root;
    cfsw_call (c);
  }
}

static void
find_commit_by_time_cb (unsigned index, bigint c, int err)
{
  if (err) {
    warn << "cannot find commit specified by the -D option\n";
    pastwatch_exit (1);
    return;
  }
  arg_tag[index] = armor64A (c);
  dispatch();
}

static void
translate_time (bigint c)
{
  // translate any -D arguments

  for (unsigned i=0; i<2; i++) {
    if (arg_date[i] && arg_date[i] != "") {
      if (c == 0) {
	warn << "-D option requires a branch\n";
	pastwatch_exit (1);
	return;
      }
      else {
	time_t t = get_date ((char *)arg_date[i].cstr ());
	if (t >= 0) {
	  find_commit_by_time (t, c, wrap (find_commit_by_time_cb, i));
	  return;
	}
	else
	  assert (0);
      }
    }
  }

  dispatch();
}

static void
got_cfsrw (cfsrw_status status)
{
  if (status) {
    warn << "Cannot read the local replica, error " << status << "\n";
    if (status == CFSRWERR_NOPROJECT) {
      warn << "  Could not find project: " << arg_projstr << "\n";
      warn << "  Does $PASTWATCH_ROOT/" << arg_projstr << "/project.id exist?\n";
    } else if (status == CFSRWERR_NOREPLICA) 
      warn << "  Could not find local replica for project: " << arg_projstr << "\n";
    pastwatch_exit (1);
    return;
  }

  if (arg_debugmode) {
    if (!anonymous)
      warn << translate_key_to_name (arg_keystr)
           << ",project:" << arg_projstr << "\n";
    else
      warn << "anonymous,project:" << arg_projstr << "\n";
  }
 
  if (!anonymous) {
    ptr<projinterface> proj = _cfsrw->getprojinf ();
    for (unsigned i=0; i<proj->nlogs (); i++)
      if (proj->get_log_ptr(i) == arg_key)
        _myname = proj->get_log_label(i);
    if (!_myname || _myname == "") {
      warn << "Could not find the specified log key (" << armor64A(arg_key) 
	   << ") in the project.\n";
      pastwatch_exit (1);
      return;
    }
  }

  const ptr<replica> r = _cfsrw->getreplica ();
  ptr<btree_srv> btree = r->btree ();
  _root = r->root ();

  //
  // translate tags
  //
  for(unsigned i=0; i<2; i++) {
    if (arg_tag[i] && arg_tag[i] != "") {
      bigint blah;
      if (!btree->lookup_by_tag (arg_tag[i], blah)) {
	warn << arg_tag[i] << " maps to zero or more than one version\n";
	pastwatch_exit (1);
	return;
      }
      arg_tag[i] = armor64A (blah);
    }
  }

  if (arg_jointag && arg_jointag != "") {
    bigint blah;
    if (!btree->lookup_by_tag (arg_jointag, blah)) {
      warn << arg_jointag << " maps zero or to more than one version\n";
      pastwatch_exit (1);
      return;
    }
    arg_jointag = armor64A (blah);
  }

  if (arg_branchfrom && arg_branchfrom != "") {
    bigint blah;
    if (!btree->lookup_by_tag (arg_branchfrom, blah)) {
      warn << arg_branchfrom << " maps zero or to more than one version\n";
      pastwatch_exit (1);
      return;
    }
    arg_branchfrom = armor64A (blah);
  }

  lock_repository(_root);
}

static void
start ()
{
  if (!anonymous)
    _cfsrw = cfsrw::alloc (arg_projid, arg_key, sk, wrap (got_cfsrw));
  else
    _cfsrw = cfsrw::alloc (arg_projid, wrap (got_cfsrw));
}

static void
nextproj3 (log_status status) 
{
  if (status) {
    if (status == LOGERR_MAKELKEYFAIL) {
      warn << "Error creating a keypair for the new log\n";
    } else if (status == LOGERR_MAKEULOGFAIL) {
      warn << "Error creating the new member log\n";
    } else if (status == LOGERR_MAKEUBLOCKFAIL) {
      warn << "Error adding the new log keypair to the member block\n";
    } else {
      warn << "Error adding the new member log\n";
    }
    pastwatch_exit (1);
    return;
  }
  
  if (arg_logkeyfile && arg_logkeyfile != "") {
    if (!(sk = read_skey (arg_logkeyfile)))
      pastwatch_exit (1);
    arg_key = ns::pkid (sk);
    arg_keystr = armor64A (arg_key);
  }
  else {
    arg_logkeyfile = "";
    anonymous = true;
  
    if (arg_command == cmd_import ||
        arg_command == cmd_commit ||
        arg_command == cmd_branch ||
        arg_command == cmd_rtag ||
        arg_command == cmd_rename ||
        arg_command == cmd_add ||
        arg_command == cmd_remove ||
        arg_command == cmd_rescue) {
      warn << "cannot run " << cmd2str (arg_command)
	   << " in anonymous mode\n";
      pastwatch_exit (1);
      return;
    }
  }
  start();
}

static void
nextproj2 ()
{
  if (arg_logkeyfile == "" && arg_userkeyfile != "") {
    warn << "Creating a new log key and log for this replica.\n";
    create_new_log(wrap (nextproj3)); // sets arg_logkeyfile
  } else {
    delaycb(0, 0, wrap(nextproj3, LOG_OK));
  }
}

static void
nextproj1 ()
{
  if (arg_projids.size () == 0) {
    _cfsrw = 0;
    nsobj = 0;
    exit (0);
  }

  files = 0;
  files = arg_paths.pop_front ();
  arg_projstr = arg_projids.pop_front ();
  arg_logkeyfile = arg_logkeys.pop_front ();
  arg_userkeyfile = arg_userkeys.pop_front ();
  arg_branch = arg_branches.pop_front ();

  if (reset_global_states ()) {
    pastwatch_exit (1);
    return;
  }

  if (!dearmor64A_bigint (arg_projstr, arg_projid)) {
    warn << "cannot parse project string: " << arg_projstr << "\n";
    pastwatch_exit (1);
    return;
  }
  nsobj = ns::alloc (arg_projid);
  nsobj->connect (wrap (nextproj2));
}

extern int getopts (int argc, char **argv);

int
main (int argc, char **argv)
{
  t0 = getusec ();
  setprogname (argv[0]);
  random_init ();
  check_create_replicadir();
  
  if (getopts (argc, argv))
    exit (1);

  sigcb (SIGHUP, wrap (pastwatch_exit, -1));
  sigcb (SIGUSR1, wrap (pastwatch_exit, -1));
  sigcb (SIGUSR2, wrap (pastwatch_exit, -1));
  sigcb (SIGALRM, wrap (pastwatch_exit, -1));
  sigcb (SIGINT, wrap (pastwatch_exit, -1));
  sigcb (SIGTERM, wrap (pastwatch_exit, -1));
  
  if (arg_projids.size () == 0) {
    warn << "no project, stopping.\n";
    exit (1);
  }

  nextproj1 ();
  amain();
}

