/*
 *  Copyright (C) 2002-2004  Massachusetts Institute of Technology
 *
 *  This program is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU General Public License as
 *  published by the Free Software Foundation; either version 2, or (at
 *  your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful, but
 *  WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *  General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307
 *  USA
 */

#include "pastwatch.h"
#include "cfsr.h"
#include "io.h"

#define CONCURRENT_READS 64

class past_read_obj : public virtual refcount
{
private:
  typedef callback<void,cfsrw_status>::ref cb_t;

  int fd;
  bigint fh;
  cb_t cb;

  uint64 size;
  uint64 off;
  unsigned int reads;
  int error;
  bigint cr_addr;
  
public:
  void done () 
  {
    if (reads == 0) {
      if (error)
	cb (CFSRWERR_IO);
      else
	cb (CFSRW_OK);
    }
  }

  void
  read_reply (uint64 off, uint64 cnt, cfsrcallp c, cfsrw_status status)
  {
    reads --;

    if (status || error) {
      error = 1;
      done ();
      return;
    }

    cfsr_read_res *res = static_cast<cfsr_read_res *> (c->res ());

    if (res->data->len != cnt) {
      warn << "asked for " << off << "+" << cnt
	   << ", got " << res->data->len << "\n";
      error = 1;
      done ();
      return;
    }

    pwrite (fd, res->data->data, res->data->len, off);
    do_read ();
    done ();
  }

  bool do_read() 
  {
    if (off >= size)
      return true;

    uint64 cnt = size - off;
    cnt = cnt > SN_BLKSIZE ? SN_BLKSIZE : cnt;
    reads ++;

    cfsrcallp c = New refcounted<cfsrcall>
      (CFSRPROC_READ, wrap (mkref(this), &past_read_obj::read_reply, off, cnt));
    cfsr_read_arg *arg = static_cast<cfsr_read_arg *> (c->arg ());
    arg->file = fh;
    arg->off = off;
    arg->cnt = cnt;
    off += cnt;

    cfsr_call (cr_addr, c);
    return false;
  }

  void get_attr (cfsrcallp c, cfsrw_status status)
  {
    if (!status) {
      cfsr_getattr_res *res = static_cast<cfsr_getattr_res *> (c->res ());
      size = res->attr->size;
      off = 0;
      bool eof = false;
      for (int i = 0; i < CONCURRENT_READS && !eof; i++)
        eof = do_read ();
    }
    else
      error = 1;
    done ();
  }

  void init ()
  {
    if (ftruncate (fd, 0) < 0) {
      error = 1;
      done ();
      return;
    }

    cfsrcallp c = New refcounted<cfsrcall>
      (CFSRPROC_GETATTR, wrap (mkref(this), &past_read_obj::get_attr));
    cfsr_getattr_arg *arg = static_cast<cfsr_getattr_arg *> (c->arg ());
    arg->object = fh;
    cfsr_call (cr_addr, c);
  }
  
  past_read_obj (int fd, bigint fh, bigint cr_addr, past_read_obj::cb_t cb)
    : fd(fd), fh(fh), cb(cb), reads(0), error(0), cr_addr(cr_addr)
  {
    init ();
  }

  ~past_read_obj() {
   }
};

void
read_file (int fd, bigint fh, bigint cr_addr,
           callback<void, cfsrw_status>::ref cb)
{
  ptr<past_read_obj> p = New refcounted<past_read_obj> (fd, fh, cr_addr, cb);
}

